# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
import os
import re
import sys


# set the platform currently being used and the major and minor Python version
PLATFORM = sys.platform
PYTHON_MAJOR = sys.version_info.major
PYTHON_MINOR = sys.version_info.minor

# set the platform type (use startswith to handle old versions with ver nums)
POSIX = os.name == 'posix'
WINDOWS = os.name == 'nt'
LINUX = sys.platform.startswith('linux')
MACOS = sys.platform.startswith('darwin')
OSX = MACOS  # deprecated alias (thanks David F for the reminder)
FREEBSD = sys.platform.startswith('freebsd')
OPENBSD = sys.platform.startswith('openbsd')
NETBSD = sys.platform.startswith('netbsd')
BSD = FREEBSD or OPENBSD or NETBSD
SUNOS = sys.platform.startswith(('sunos', 'solaris'))
AIX = sys.platform.startswith('aix')

# get the users home path for storing files
if POSIX:
    HOMEDIR = os.path.expanduser('~')
elif WINDOWS:
    HOMEDIR = os.environ['APPDATA']
else:
    # unsupported operating system
    HOMEDIR = os.path.realpath(os.sep)

# set the storage constant
if MACOS:
    APP_SUP = os.path.join(HOMEDIR, 'Library', 'Application Support')
    STORAGE = os.path.join(APP_SUP, 'Area28', 'core')
elif WINDOWS:
    STORAGE = os.path.join(HOMEDIR, 'Area28', 'core')
elif POSIX:
    STORAGE = os.path.join(HOMEDIR, '.config', 'area28', 'core')
else:
    STORAGE = os.path.join(HOMEDIR, '.area28', 'core')


def message(msg=''):
    """Display a message to the user."""
    print(msg)


def confirm(message='confirm?'):
    """Ask the user to confirm that ALL files and directories should be deleted
    from the directory specified by the storage variable. The STORAGE variable
    changes depending on the operating system used.
    """
    confirm = ['y', 'Y', 'yes', 'ok', 'sure', 'alrighty', 'of course', 'yup']
    return bool(input(message) in confirm)


def valid_string(value, min_length=3, max_length=64):
    """Check if a string contains only alphanumeric, hyphen, and underscore
    characters and limited to 64 characters.
    """
    check = rf'^[\w_-]{{{min_length},{max_length}}}$'
    return re.match(check, value) is not None



# flake8: noqa
import subprocess  # noqa: I005 S404


def capture(command):
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_a28_no_param():
    command = ['a28']
    out, err, exitcode = capture(command)
    assert exitcode == 2
    assert out == b''
    message = b'usage: a28 [-h] [-v] {package,pkg,system,sys} ...'
    assert err[0:len(message)] == message


def test_a28_version_param():
    command = ['a28', '-v']
    out, err, exitcode = capture(command)
    assert exitcode == 0
    message = b'a28 version '
    assert out[0:len(message)] == message
    message = b''
    assert err[0:len(message)] == message



# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
import hashlib
import json
import os
import shutil
import sys
import zipfile
from uuid import uuid4
from pkg_resources import parse_version as get_version
from . import log, utils


# try import the requests module
try:
    import requests
except ImportError:
    log.fatal('the requests module is required, please pip install requests')
    sys.exit(1)


def options(mainparser):
    """Argparse options added to the cli."""
    parser = mainparser.add_parser(
        'package',
        aliases=['pkg'],
        help='package actions'
    )
    subparser = parser.add_subparsers(
        dest='package',
        required=True,
        help='package',
    )
    parser_init = subparser.add_parser(
        'init',
        help='Initialize a package',
    )
    parser_init.set_defaults(func=initialize)
    parser_init.add_argument(
        'path',
        default='.',
        help='package path default \'.\'',
    )
    parser_init.add_argument(
        '-s',
        '--scope',
        required=True,
        help='package scope eg. @group-name',
    )
    parser_init.add_argument(
        '-n',
        '--name',
        required=True,
        help='package mnamespace eg. pkg-name',
    )
    parser_init.add_argument(
        '-t',
        '--type',
        required=True,
        choices=[
            'api', 'app', 'chat', 'event',
            'interface', 'logger', 'metadata',
            'preferences', 'realtime', 'repository',
            'ui', 'units'
        ],
        help='package type',
    )
    parser_init.add_argument(
        '--bin',
        action='store_true',
        help='create the bin directory',
    )
    parser_init.add_argument(
        '--script',
        action='store_true',
        help='create the scripts directory',
    )
    parser_init.add_argument(
        '-f',
        '--force',
        action='store_true',
        help='force overwriting package.json',
    )
    parser_build = subparser.add_parser(
        'build',
        help='build package',
    )
    parser_build.set_defaults(func=build)
    parser_build.add_argument(
        '--src',
        required=True,
        help='package source directory',
    )
    parser_build.add_argument(
        '--dest',
        default='',
        help='destination directory',
    )
    parser_install = subparser.add_parser(
        'install',
        help='install package',
    )
    parser_install.set_defaults(func=install)
    parser_install.add_argument(
        '--pkg',
        required=True,
        help='package a28 package file',
    )


def initialize(args):
    """Initialize a package in a specified directory."""
    (path, name, scope, schema) = _get_init_args(args)
    log.info(f'initializing @{scope}/{name} in {path}')
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)
    subdirs = ['extensions', 'plugins']
    json_file = os.path.join(path, 'package.json')
    exists = os.path.isfile(json_file)
    json_data = {}
    json_data['name'] = f'@{scope}/{name}-{schema}'
    json_data['version'] = '0.1.0'
    if args.bin:
        log.debug('adding bin')
        subdirs.append('bin')
        json_data['bin'] = {}
    if args.script:
        log.debug('adding scripts')
        subdirs.append('script')
        json_data['scripts'] = {}
    if args.type == 'application':
        subdirs.append('plugin')
        _build_structure(path, subdirs)
    message = 'overwrite package.json?'
    if not exists or args.force or utils.confirm(message):
        log.info('creating package.json')
        with open(json_file, 'w') as pkg_file:
            json.dump(json_data, pkg_file, indent=4)
    else:
        log.info('not overwriting package.json')


def build(args):
    """Build package."""
    src = args.src
    dest = args.dest
    meta = get_info(src)
    pkg = package(src, dest, meta)
    utils.message(pkg)


def install(args):
    """install / update the local package."""
    pkg_hash = generate_hash(args.pkg)
    meta = extract_meta(args.pkg)
    meta['hash'] = pkg_hash
    install_local(args.pkg, meta)
    utils.message('installed {}'.format(meta['name']))


def extract_meta(pkg):
    """Extract the meta information from a packaged package."""
    with zipfile.ZipFile(pkg, 'r') as z:
        package = next(x for x in z.namelist() if x.endswith('package.json'))
        with z.open(package) as manifest:
            return get_info(manifest, True)


def get_info(src, extracted=False):
    """Get information about a package by using a combination of the data in
    the src package.json file and the manifest.json file.
    """
    if extracted:
        data = json.load(src)
    else:
        package = os.path.join(src, 'package.json')
        with open(package) as package_data:
            data = json.load(package_data)

    # load the manifest data
    manifest = os.path.join(utils.STORAGE, 'manifest.json')
    with open(manifest) as manifest_data:
        m_data = json.load(manifest_data)

    # copy data from the manifest package if exists
    if data['name'] in m_data['packages']:
        m_package = m_data['packages'][data['name']]
        data['identifier'] = m_package['identifier']
    else:
        data['identifier'] = str(uuid4())
    return data


def install_local(pkg, meta):
    """Install a package locally by updating the manifest.json without
    touching the etag.json file. This will be overwritten when a new official
    manifest.json is published."""
    dest = os.path.join(utils.STORAGE, 'cache')
    shutil.copy(pkg, dest)
    manifest = os.path.join(utils.STORAGE, 'manifest.json')
    with open(manifest, 'r+') as manifest_data:
        data = json.load(manifest_data)
        data['packages'][meta['name']] = meta
        if meta['name'] not in data['required']:
            data['required'].append(meta['name'])
        manifest_data.seek(0)
        json.dump(data, manifest_data, indent=4)
        manifest_data.truncate()


def package(src, dest, meta):
    """Build an a28 package from the provided src directory. The package will
    be saved to the dest directory. A package needs to be provided containing
    at least an identifier and a version number."""
    version = meta['version']
    identifier = meta['identifier']
    filename = '{}-{}.{}'.format(identifier, version, 'a28')
    filename = os.path.join(dest, filename)

    if not os.path.exists(dest):
        os.makedirs(dest)

    a28 = zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED)
    exclude = (['build', '.vscode'])
    for root, dirs, files in os.walk(src, topdown=True):
        dirs[:] = [d for d in dirs if d not in exclude]
        for current in files:
            i_file = os.path.join(root, current)
            fl = os.path.relpath(
                os.path.join(root, current),
                os.path.join(src, '..', '..'),
            )
            a28.write(i_file, fl)
    a28.close()
    return filename


def fetch(url='https://packages.a28.io/manifest.json', dest=None):
    """Fetch the manifest from it's online source."""
    etag = os.path.join(utils.STORAGE, 'manifest.etag')
    request = requests.head(url, allow_redirects=True)
    latest = request.headers['ETag']
    request.close()

    with open(etag, 'w+') as file:
        current = file.read().replace('\n', '')
        if get_version(latest) > get_version(current):
            file.write(latest)
            request = requests.get(url, allow_redirects=True)
            with open(dest, 'wb') as manifest:
                manifest.write(request.content)
            request.close()


def generate_hash(pkg):
    """Generate a hash for a package."""
    sha256_hash = hashlib.sha256()
    with open(pkg, 'rb') as f:
        # read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b''):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def prebuild(src):
    """Execute any pre-build scripts. Usually used to create installable
    tarballs for applications.
    """


def _get_init_args(args):
    path = os.path.realpath(args.path)
    if not utils.valid_string(args.name):
        raise ValueError(f'{args.name} is not a valid name')
    name = args.name.lower()
    if not utils.valid_string(args.scope):
        raise ValueError(f'{args.scope} is not a valid scope')
    scope = args.scope.lower()
    if not utils.valid_string(args.type, min_length=2):
        raise ValueError(f'{args.type} is not a valid name')
    schema = args.type.lower()
    return (path, name, scope, schema)


def _build_structure(path, subdirs):
    for subdir in subdirs:
        os.makedirs(os.path.join(path, subdir), exist_ok=True)



# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
import os
import shutil
from . import utils


def options(mainparser):
    """Argparse options added to the cli."""
    parser = mainparser.add_parser(
        'system',
        aliases=['sys'],
        help='system actions'
    )
    subparser = parser.add_subparsers(
        dest='system',
        required=True,
        help='system',
    )
    parser_exist = subparser.add_parser(
        'exists',
        help='check if the system configuration exists.',
    )
    parser_exist.set_defaults(func=exists)
    parser_clean = subparser.add_parser(
        'clean',
        help='clean (delete) the configuration permanently.',
    )
    parser_clean.set_defaults(func=clean)
    parser_clean.add_argument(
        '-f',
        '--force',
        action='store_true',
        help='force the configuration to be removed bypassing confirmation',
    )


def clean(args):
    """Using the shutil utilities, delete all the files provided in the STORAGE
    directory. If an error occurs, catch the exception and print it out to the
    STDOUT.
    """
    message = f"Delete ALL configuration data from '{utils.STORAGE}'? y/n?"
    if not os.path.isdir(utils.STORAGE):
        utils.message('No configuration to clean.')
        return

    elif not args.force and not utils.confirm(message):
        utils.message(f'Not deleting {utils.STORAGE}')
        return

    try:
        shutil.rmtree(utils.STORAGE)
        utils.message(f'Deleted all files and directories in {utils.STORAGE}')
    except OSError as e:
        utils.message(f'Error: {utils.STORAGE} : {e.strerror}')


def exists(_):
    """Check if the configuration directory exists."""
    if os.path.isdir(utils.STORAGE):
        utils.message(f"Configuration exists at '{utils.STORAGE}'.")
    else:
        utils.message(f"No configuration exists at '{utils.STORAGE}'.")



# flake8: noqa
import subprocess  # noqa: I005 S404


def capture(command):
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_a28_package_no_param():
    command = ['a28', 'package']
    out, err, exitcode = capture(command)
    assert exitcode == 2
    assert out == b''
    message = b'usage: a28 package [-h] {init,build,install} ...'
    assert err[0:len(message)] == message


def test_a28_init_no_param():
    command = ['a28', 'package', 'init']
    out, err, exitcode = capture(command)
    assert exitcode == 2
    assert out == b''
    message = b'usage: a28 package init [-h] -s SCOPE -n NAME -t'
    assert err[0:len(message)] == message


def test_a28_build_no_param():
    command = ['a28', 'package', 'build']
    out, err, exitcode = capture(command)
    assert exitcode == 2
    assert out == b''
    message = b'usage: a28 package build [-h] --src SRC [--dest DEST]'
    assert err[0:len(message)] == message


def test_a28_install_no_param():
    command = ['a28', 'package', 'install']
    out, err, exitcode = capture(command)
    assert exitcode == 2
    assert out == b''
    message = b'usage: a28 package install [-h] --pkg PKG'
    assert err[0:len(message)] == message



#!/usr/bin/env python
# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
#    _                ___ ___   _____       _             _           _
#   /_\  _ _ ___ __ _|_  | _ ) |_   _|__ __| |_  _ _  ___| |___  __ _(_)___ ___
#  / _ \| '_/ -_) _` |/ // _ \   | |/ -_) _| ' \| ' \/ _ \ / _ \/ _` | / -_|_-<
# /_/ \_\_| \___\__,_/___\___/   |_|\___\__|_||_|_||_\___/_\___/\__, |_\___/__/
#                                                               |___/
"""Build packages to be submitted to Area28."""
import argparse
import sys
from . import __version__, package, system


def main(args=None):
    """Main command line entry point."""
    args = args if args else sys.argv[1:]
    parser = argparse.ArgumentParser(
        'a28',
        description='Area28 development kit'
    )
    app = '%(prog)s version ' + __version__
    parser.add_argument('-v', '--version', action='version', version=app)
    sub = parser.add_subparsers(dest='action', required=True, help='actions')
    package.options(sub)
    system.options(sub)
    args = parser.parse_args(args)
    args.func(args)



# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
#    _                ___ ___   _____       _             _           _
#   /_\  _ _ ___ __ _|_  | _ ) |_   _|__ __| |_  _ _  ___| |___  __ _(_)___ ___
#  / _ \| '_/ -_) _` |/ // _ \   | |/ -_) _| ' \| ' \/ _ \ / _ \/ _` | / -_|_-<
# /_/ \_\_| \___\__,_/___\___/   |_|\___\__|_||_|_||_\___/_\___/\__, |_\___/__/
#                                                               |___/
"""Build packages to be submitted to Area28."""
import logging


# current version
__version__ = '0.3.0'

# enable logging
logging.basicConfig()
log = logging.getLogger()
log.setLevel(logging.DEBUG)

# try use colored logs
try:
    import coloredlogs

    coloredlogs.install(level='DEBUG', logger=log)
except ImportError:
    coloredlogs = None
    log.warning('for better logging, please pip install coloredlogs')



# flake8: noqa
import subprocess  # noqa: I005 S404


def capture(command):
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_a28_system_no_param():
    command = ['a28', 'system']
    out, err, exitcode = capture(command)
    assert exitcode == 2
    assert out == b''
    message = b'usage: a28 system [-h] {exists,clean} ...'
    assert err[0:len(message)] == message


def test_a28_exists_no_param():
    command = ['a28', 'system', 'exists']
    out, err, exitcode = capture(command)
    assert exitcode == 0
    message = b'No configuration exists at '
    assert out[0:len(message)] == message


def test_a28_clean_no_param():
    command = ['a28', 'system', 'clean']
    out, err, exitcode = capture(command)
    assert exitcode == 0
    message = b'No configuration to clean.'
    assert out[0:len(message)] == b'No configuration to clean.'
